/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 *************************************************************************************
 * \file header.c
 *
 * \brief
 *    H.26L Slice and Sequence headers
 *
 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *      - Stephan Wenger                  <stewe@cs.tu-berlin.de>
 *************************************************************************************
 */

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "elements.h"
#include "header.h"
#include "rtp.h"
#include "mbuffer.h"
#include "encodeiff.h"
#include "defines.h"

// A little trick to avoid those horrible #if TRACE all over the source code
#if TRACE
#define SYMTRACESTRING(s) strncpy(sym->tracestring,s,TRACESTRING_SIZE)
#else
#define SYMTRACESTRING(s) // do nothing
#endif


/*!
 ********************************************************************************************
 * \brief 
 *    Write a slice header
 ********************************************************************************************
*/
int SliceHeader()
{
  int dP_nr = assignSE2partition[input->partition_mode][SE_HEADER];
  Bitstream *currStream = ((img->currentSlice)->partArr[dP_nr]).bitstream;
  DataPartition *partition = &((img->currentSlice)->partArr[dP_nr]);
  SyntaxElement *sym;
  int len = 0;
  int Quant = img->qp;
  int i,j;

  if ((sym=(SyntaxElement*)calloc(1,sizeof(SyntaxElement)))==NULL) no_mem_exit("SliceHeader:sym");

  sym->type = SE_HEADER;                 // This will be true for all symbols generated here


// Note: the pic_parameter_set_id MUST be the first item of the slice header
// not only because it's in the standard, but also because the decoder
// determines the parset to be used for decoding from the first ue(v) coded
// symbol of the slice header!  Don't add anything here!!!

//! Note: Encoder supports only one pic/seq parameter set, hence value is
//! hard coded to zero

  sym->mapping = n_linfo2;               // change to unsigned integer
  sym->value1 = 0;                       // ue(v)
  SYMTRACESTRING("SH: pic_parameter_set_id");
  len += writeSyntaxElement_UVLC (sym, partition);
  

  // frame_num
  assert (img->frame_num < 1<<(LOG2_MAX_FRAME_NUM_MINUS4+4));  // check that it fits
  sym->len = LOG2_MAX_FRAME_NUM_MINUS4 + 4;
  sym->bitpattern = img->frame_num;
  SYMTRACESTRING("SH: frame_num");
  len += writeSyntaxElement_fixed (sym, partition);

  //field_pic_flag    u(1)
  if(img->pstruct ==TOP_FIELD || img->pstruct ==BOTTOM_FIELD)
    sym->bitpattern = 1;                 
  else 
    sym->bitpattern = 0;
  if(sym->bitpattern != img->fld_flag)
  {
    //error("img->fld_flag seems wrong", -1005);
    //img->fld_flag is wrong on first field in 5.0
  }
  sym->len = 1;
  SYMTRACESTRING("SH: field_pic_flag");
  len += writeSyntaxElement_fixed (sym, partition);
  
  if(img->pstruct ==TOP_FIELD || img->pstruct ==BOTTOM_FIELD){
    //bottom_field_flag
    if(img->pstruct == BOTTOM_FIELD)sym->bitpattern = 1;                  
    else sym->bitpattern = 0;   
    sym->len = 1;
    SYMTRACESTRING("SH: bottom_field_flag");
    len += writeSyntaxElement_fixed (sym, partition);
  }

  //! Note: not yet an u(v) -- wait for outcome of Away meeting
  sym->mapping = n_linfo2;               // unsigned integer, ue(v)
  assert (img->current_mb_nr < (1<<15));
  SYMTRACESTRING("SH first_mb_in_slice");
  sym->value1 = img->current_mb_nr;
  len += writeSyntaxElement_UVLC (sym, partition);

  //! Note: this is still the old slice type, which distinguishes between
  //! single and multipred for P and B frames.  Too scary to change this
  //! just now.  I believe, the decoder wouldn't care if multpred is
  //! always signalled -- but to check it out, please don't forget to
  //! change the encoder config file so that only one refframe is used.

  // Slice type
  sym->mapping = n_linfo2;               // unsigned integer, ue(v)
  select_picture_type (sym);
  SYMTRACESTRING("SH slice_type (old)");
  len+= writeSyntaxElement_UVLC (sym, partition);

  if (img->pic_order_cnt_type < 2 && !img->delta_pic_order_always_zero_flag)
  {
    sym->mapping = dquant_linfo;               // change to signed integer
    sym->value1 = img->delta_pic_order_cnt[0];    //  se(v)  //same value for both fields
    SYMTRACESTRING("delta_pic_order_cnt[0]");
    len += writeSyntaxElement_UVLC (sym, partition);

    if (img->pic_order_present_flag && !img->fld_flag)
    {
      assert ("This case is not yet implemented\n");
      assert (0==1);
    }
  }
  else
  {
    assert ("This case is not implemented\n");
    assert (0==1);
  }
  

  //! redundant slice info redundant_pic_cnt is missing here


  // Direct Mode Type selection for B pictures
  if (img->type==B_IMG || img->type==BS_IMG)
  {
    SYMTRACESTRING("SH direct_spatial_mv_pred_flag");
    sym->bitpattern = input->direct_type;  // 1 for Spatial Direct, 0 for temporal Direct
    sym->len = 1;
    len += writeSyntaxElement_fixed(sym, partition);
  }


  sym->mapping = n_linfo2;               // change to unsigned integer
  
  toprefpoc[0]=toprefpoc[0];
  bottomrefpoc[0]=bottomrefpoc[0];



  // picture structure, (0: progressive, 1: top field, 2: bottom field, 
  // 3: top field first, 4: bottom field first)
  sym->value1 = img->pstruct;
  SYMTRACESTRING("SH Picture Stucture");
  len += writeSyntaxElement_UVLC (sym, partition);

  // Finally, write Reference Picture ID 
  // WYK: Oct. 16, 2001. Now I use this for the reference frame ID (non-B frame ID). 
  // If current frame is a B-frame, it will not change; otherwise it is increased by 1 based
  // on refPicID of the previous frame. Thus, the decoder can know how many  non-B 
  // frames are lost, and then can adjust the reference frame buffers correctly.
  if (1)
  {
    sym->value1 = img->refPicID%16;         // refPicID, variable length
 
    SYMTRACESTRING("SH RefPicID");
    len += writeSyntaxElement_UVLC (sym, partition);
  }

  // NOTE: following syntax elements have to be moved into picture_layer_rbsp()
  // Not really.  They keep changing during the startup of the sequence (and
  // likely also after an IDR).  WHat is missing is the bit that signals their
  // presence.
  //! The above needs to be implemented!

  if (img->type==INTER_IMG || img->type==B_IMG || img->type==BS_IMG)
  {
    sym->value1 = img->num_ref_pic_active_fwd_minus1;
    SYMTRACESTRING("num_ref_pic_active_fwd_minus1");
    len += writeSyntaxElement_UVLC (sym, partition);
    if (img->type==B_IMG || img->type==BS_IMG)
    {
      sym->value1 = img->num_ref_pic_active_bwd_minus1;
      SYMTRACESTRING("num_ref_pic_active_bwd_minus1");
      len += writeSyntaxElement_UVLC (sym, partition);
    }
  }



  SYMTRACESTRING("SH SliceQuant");
  sym->mapping = dquant_linfo;  
  sym->value1 = Quant - (MAX_QP - MIN_QP +1)/2;
  len += writeSyntaxElement_UVLC (sym, partition);
  if (img->types==SP_IMG )  
  {
    if (img->types==SP_IMG && img->type!=INTRA_IMG) // Switch Flag only for SP pictures
    {
      SYMTRACESTRING("SH SWITCH FLAG");
      sym->bitpattern = 0;  // 1 for switching SP, 0 for normal SP
      sym->len = 1;
      len += writeSyntaxElement_fixed(sym, partition);
    }
    
    SYMTRACESTRING("SH SP SliceQuant");
    sym->value1 = img->qpsp - (MAX_QP - MIN_QP +1)/2;
    len += writeSyntaxElement_UVLC (sym, partition);
  }

  if (input->LFSendParameters)
  {
    SYMTRACESTRING("SH LF_DISABLE FLAG");
    sym->bitpattern = input->LFDisable;  // Turn loop filter on/off on slice basis 
    sym->len = 1;
    len += writeSyntaxElement_fixed(sym, partition);

    if (!input->LFDisable)
    {
      sym->mapping = dquant_linfo;           // Mapping rule: Signed integer
      SYMTRACESTRING("SH LFAlphaC0OffsetDiv2");
      sym->value1 = input->LFAlphaC0Offset>>1; // Convert from offset to code 
      len += writeSyntaxElement_UVLC (sym, partition);

      SYMTRACESTRING("SH LFBetaOffsetDiv2");
      sym->value1 = input->LFBetaOffset>>1; // Convert from offset to code 
      len += writeSyntaxElement_UVLC (sym, partition);
    }
  }

  // putting pred_weight_table() here
  if ((img->type == INTER_IMG && input->WeightedPrediction) || 
          ((img->type == B_IMG || img->type == BS_IMG) && input->WeightedBiprediction == 1))
  {
    sym->mapping = dquant_linfo;               // change to signed integer

    sym->value1 = luma_log_weight_denom;  //    se(v)   
    SYMTRACESTRING("luma_log_weight_denom");
    len += writeSyntaxElement_UVLC (sym, partition);
  
    sym->value1 = chroma_log_weight_denom;  //    se(v)   
    SYMTRACESTRING("chroma_log_weight_denom");
    len += writeSyntaxElement_UVLC (sym, partition);

    for (i=0; i<= img->num_ref_pic_active_fwd_minus1; i++)
    {
      if (wp_weight[0][i][0] != 1<<luma_log_weight_denom)
      {
        sym->mapping = n_linfo2;               // change to unsigned integer
        sym->len = 1;
        sym->bitpattern = 1;   
        SYMTRACESTRING("luma_weight_flag_l0");
        len += writeSyntaxElement_fixed (sym, partition);

        sym->mapping = dquant_linfo;               // change to signed integer
        sym->value1 = wp_weight[0][i][0];  //    se(v)   
        SYMTRACESTRING("luma_weight_l0");
        len += writeSyntaxElement_UVLC (sym, partition);

        sym->value1 = wp_offset[0][i][0];  //    se(v)   
        SYMTRACESTRING("luma_offset_l0");
        len += writeSyntaxElement_UVLC (sym, partition);
      }
      else
      {
        sym->mapping = n_linfo2;               // change to unsigned integer
        sym->len = 1;
        sym->bitpattern = 0;   
        SYMTRACESTRING("luma_weight_flag_l0");
        len += writeSyntaxElement_fixed (sym, partition);
      }

      for (j=1; j<3; j++)
      {
        if (wp_weight[0][i][j] != 1<<chroma_log_weight_denom)
        {
          sym->mapping = n_linfo2;               // change to unsigned integer
          sym->len = 1;
          sym->bitpattern = 1;   
          SYMTRACESTRING("chroma_weight_flag_l0");
          len += writeSyntaxElement_fixed (sym, partition);

          sym->mapping = dquant_linfo;               // change to signed integer
          sym->value1 = wp_weight[0][i][j];  //    se(v)   
          SYMTRACESTRING("chroma_weight_l0");
          len += writeSyntaxElement_UVLC (sym, partition);

          sym->value1 = wp_offset[0][i][j];  //    se(v)   
          SYMTRACESTRING("chroma_offset_l0");
          len += writeSyntaxElement_UVLC (sym, partition);
        }
        else
        {
          sym->mapping = n_linfo2;               // change to unsigned integer
          sym->len = 1;
          sym->bitpattern = 0;   
          SYMTRACESTRING("chroma_weight_flag_l0");
          len += writeSyntaxElement_fixed (sym, partition);
        }
      }
    }
  }
  if ((img->type == B_IMG || img->type == BS_IMG) && input->WeightedBiprediction == 1)
  {
    for (i=0; i<= img->num_ref_pic_active_bwd_minus1; i++)
    {
      if (wp_weight[1][i][0] != 1<<luma_log_weight_denom)
      {
        sym->mapping = n_linfo2;               // change to unsigned integer
        sym->len = 1;
        sym->bitpattern = 1;   
        SYMTRACESTRING("luma_weight_flag_l1");
        len += writeSyntaxElement_fixed (sym, partition);

        sym->mapping = dquant_linfo;               // change to signed integer
        sym->value1 = wp_weight[1][i][0];  //    se(v)   
        SYMTRACESTRING("luma_weight_l1");
        len += writeSyntaxElement_UVLC (sym, partition);

        sym->value1 = wp_offset[1][i][0];  //    se(v)   
        SYMTRACESTRING("luma_offset_l1");
        len += writeSyntaxElement_UVLC (sym, partition);
      }
      else
      {
        sym->mapping = n_linfo2;               // change to unsigned integer
        sym->len = 1;
        sym->bitpattern = 0;   
        SYMTRACESTRING("luma_weight_flag_l1");
        len += writeSyntaxElement_fixed (sym, partition);
      }

      for (j=1; j<3; j++)
      {
        if (wp_weight[1][i][j] != 1<<chroma_log_weight_denom)
        {
          sym->mapping = n_linfo2;               // change to unsigned integer
          sym->len = 1;
          sym->bitpattern = 1;   
          SYMTRACESTRING("chroma_weight_flag_l1");
          len += writeSyntaxElement_fixed (sym, partition);

          sym->mapping = n_linfo2;               // change to unsigned integer
          sym->value1 = wp_weight[1][i][j];  //    se(v)   
          SYMTRACESTRING("chroma_weight_l1");
          len += writeSyntaxElement_UVLC (sym, partition);

          sym->value1 = wp_offset[1][i][j];  //    se(v)   
          SYMTRACESTRING("chroma_offset_l1");
          len += writeSyntaxElement_UVLC (sym, partition);
        } 
        else
        {
          sym->mapping = n_linfo2;               // change to unsigned integer
          sym->len = 1;
          sym->bitpattern = 0;   
          SYMTRACESTRING("chroma_weight_flag_l1");
          len += writeSyntaxElement_fixed (sym, partition);
        }
      }
    }
  }
  
  len+=writeERPS(sym, partition);
  
  free(sym);

  return len;
}

/*!
 ********************************************************************************************
 * \brief 
 *    writes the ERPS syntax elements
 *
 * \return
 *    number of bits used for the ERPS
 ********************************************************************************************
*/
int writeERPS(SyntaxElement *sym, DataPartition *partition)
{
  int len=0;

#ifdef _CHECK_MULTI_BUFFER_1_
  RMPNIbuffer_t *r;
#endif

  /* RPSF: Reference Picture Selection Flags */
  SYMTRACESTRING("SH: Reference Picture Selection Flags");
  sym->value1 = 0;
  len += writeSyntaxElement_UVLC (sym, partition);

  /* PN: Picture Number */
  SYMTRACESTRING("SH: Picture Number");
  sym->value1 = img->pn;
#ifdef FLD0
  sym->value1 =0;
#endif
  len += writeSyntaxElement_UVLC (sym, partition);

#ifdef _CHECK_MULTI_BUFFER_1_

  /* RPSL: Reference Picture Selection Layer */
  SYMTRACESTRING("SH: Reference Picture Selection Layer");
  sym->value1 = 1;
  len += writeSyntaxElement_UVLC (sym, partition);

  if(img->type!=INTRA_IMG)
  {
    // let's mix some reference frames
    if ((img->pn==5)&&(img->type==INTER_IMG))
    {
      r = (RMPNIbuffer_t*)calloc (1,sizeof(RMPNIbuffer_t));
      r->RMPNI=0;
      r->Data=2;
      r->Next=NULL;
      img->currentSlice->rmpni_buffer=r;

    
      // negative ADPN follows
      SYMTRACESTRING("SH: RMPNI");
      sym->value1 = 0;
      len += writeSyntaxElement_UVLC (sym, partition);

      // ADPN
      SYMTRACESTRING("SH: ADPN");
      sym->value1 = 2;
      len += writeSyntaxElement_UVLC (sym, partition);

    }

    // End loop
    SYMTRACESTRING("SH: RMPNI");
    sym->value1 = 3;
    len += writeSyntaxElement_UVLC (sym, partition);
  }
  reorder_mref(img);
  
#else
  /* RPSL: Reference Picture Selection Layer */
  SYMTRACESTRING("SH: Reference Picture Selection Layer");
  sym->value1 = 0;
  len += writeSyntaxElement_UVLC (sym, partition);

#endif

#ifdef _CHECK_MULTI_BUFFER_2_

  SYMTRACESTRING("SH: Reference Picture Bufering Type");
  sym->value1 = 1;
  len += writeSyntaxElement_UVLC (sym, partition);


  // some code to check operation
  if ((img->pn==3) && (img->type==INTER_IMG))
  {

    // check in this frame as long term picture
//    if (img->max_lindex==0)
    {
      // set long term buffer size = 2
      SYMTRACESTRING("SH: MMCO Specify Max Long Term Index");
      // command
      sym->value1 = 4;
      len += writeSyntaxElement_UVLC (sym, partition);
      // size = 2+1 (MLP1)
      sym->value1 = 2+1;
      len += writeSyntaxElement_UVLC (sym, partition);

      img->max_lindex=2;
    }

    // assign a long term index to actual frame
    SYMTRACESTRING("SH: MMCO Assign Long Term Index to a Picture");
    // command
    sym->value1 = 3;
    len += writeSyntaxElement_UVLC (sym, partition);
    // DPN=0 for actual frame 
    sym->value1 = 0;
    len += writeSyntaxElement_UVLC (sym, partition);
    //long term ID
    sym->value1 = img->lindex;
    len += writeSyntaxElement_UVLC (sym, partition);

    // assign local long term
    init_long_term_buffer(2,img);
    init_mref(img);
    init_Refbuf(img);

    assign_long_term_id(3,img->lindex,img);
    
    img->lindex=(img->lindex+1)%img->max_lindex;


  } 
  if ((img->pn==4) && (img->type==INTER_IMG))
  {
    // delete long term picture again
    SYMTRACESTRING("SH: MMCO Mark a Long-Term Picture as Unused");
    // command
    sym->value1 = 2;
    len += writeSyntaxElement_UVLC (sym, partition);
    SYMTRACESTRING("SH: MMCO LPIN");
    // command
    sym->value1 = (img->max_lindex+img->lindex-1)%img->max_lindex;
    len += writeSyntaxElement_UVLC (sym, partition);
  } 

  // end MMCO loop
  SYMTRACESTRING("SH: end loop");
  sym->value1 = 0;
  len += writeSyntaxElement_UVLC (sym, partition);
#else
    /* RPBT: Reference Picture Bufering Type */
    SYMTRACESTRING("SH: Reference Picture Bufering Type");
    sym->value1 = 0;
    len += writeSyntaxElement_UVLC (sym, partition);
#endif 

  return len;
}




/********************************************************************************************
 ********************************************************************************************
 *
 * Local Support Functions
 *
 ********************************************************************************************
 ********************************************************************************************/



// StW Note: This function is a hack.  It would be cleaner if the encoder maintains
// the picture type in the given format.  Note further that I have yet to understand
// why the encoder needs to know whether a picture is predicted from one or more
// reference pictures.

/*!
 ************************************************************************
 * \brief
 *    Selects picture type and codes it to symbol
 ************************************************************************
 */
void select_picture_type(SyntaxElement *symbol)
{
  int multpred;

#ifdef _ADDITIONAL_REFERENCE_FRAME_
  if (input->no_multpred <= 1 && input->add_ref_frame == 0)
#else
  if (input->no_multpred <= 1)
#endif
    multpred=FALSE;
  else
    multpred=TRUE;               // multiple reference frames in motion search

  if (img->type == INTRA_IMG)
  {
    symbol->len=3;
    symbol->inf=1;
    symbol->value1 = 2;
  }
  else if((img->type == INTER_IMG) && (multpred == FALSE) ) // inter single reference frame
  {
    symbol->len=1;
    symbol->inf=0;
    symbol->value1 = 0;
  }
  else if((img->type == INTER_IMG) && (multpred == TRUE)) // inter multiple reference frames
  {
    symbol->len=3;
    symbol->inf=0;
    symbol->value1 = 1;
  }
  else if((img->type == B_IMG) && (multpred == FALSE))
  {
    symbol->len=5;
    symbol->inf=0;
    symbol->value1 = 3;
  }
  else if((img->type == B_IMG) && (multpred == TRUE))
  {
    symbol->len=5;
    symbol->inf=1;
    symbol->value1 = 4;
  }
  else if((img->type == BS_IMG) && (multpred == TRUE))
  {
    symbol->len=5;
    symbol->inf=1;
    symbol->value1 = 4;
  }
  else
  {
    error("Picture Type not supported!",1);
  }

  if((img->types == SP_IMG ) && (multpred == FALSE))
  {
    symbol->len=5;
    symbol->inf=0;
    symbol->value1 = 5;
  }
  else if((img->types == SP_IMG) && (multpred == TRUE))
  {
    symbol->len=5;
    symbol->inf=0;
    symbol->value1 = 6;
  }


#if TRACE
  snprintf(symbol->tracestring, TRACESTRING_SIZE, "Image type = %3d ", img->type);
#endif
  symbol->type = SE_PTYPE;
}



/*!
 *****************************************************************************
 *
 * \brief 
 *    int Partition_BC_Header () write the Partition type B, C header
 *
 * \return
 *    Number of bits used by the partition header
 *
 * \para Parameters
 *    PartNo: Partition Number to which the header should be written
 *
 * \para Side effects
 *    Partition header as per VCEG-N72r2 is written into the appropriate 
 *    partition bit buffer
 *
 * \para Limitations/Shortcomings/Tweaks
 *    The current code does not support the change of picture parameters within
 *    one coded sequence, hence there is only one parameter set necessary.  This
 *    is hard coded to zero.
 *
 * \para
 *    Note: this is still the old Partition_BC_Header and not yet compliant with
 *    the spec.
 *   
 * \date
 *    October 24, 2001
 *
 * \author
 *    Stephan Wenger   stewe@cs.tu-berlin.de
 *****************************************************************************/


int Partition_BC_Header(int PartNo)
{
  DataPartition *partition = &((img->currentSlice)->partArr[PartNo]);
  SyntaxElement symbol, *sym = &symbol;

  int len = 0;

  assert (input->of_mode == PAR_OF_RTP);
  assert (PartNo > 0 && PartNo < img->currentSlice->max_part_nr);

  sym->type = SE_HEADER;         // This will be true for all symbols generated here
  sym->mapping = n_linfo2;       // Mapping rule: Simple code number to len/info


  SYMTRACESTRING("RTP-PH: Picture ID");
  sym->value1 = img->currentSlice->picture_id;
  len += writeSyntaxElement_UVLC (sym, partition);

  SYMTRACESTRING("RTP-PH: Slice ID");
  sym->value1 = img->current_slice_nr;
  len += writeSyntaxElement_UVLC (sym, partition);

  return len;
}
