/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 **************************************************************************************
 * \file
 *    slice.c
 * \brief
 *    generate the slice header, setup the bit buffer for slices,
 *    and generates the slice NALU(s)

 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *      - Thomas Stockhammer            <stockhammer@ei.tum.de>
 *      - Detlev Marpe                  <marpe@hhi.de>
 *      - Stephan Wenger                <stewe@cs.tu-berlin.de>
 ***************************************************************************************
 */

#include "contributors.h"

#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "header.h"
#include "rtp.h"
#include "encodeiff.h"
#include "sei.h"
#include "nalu.h"
#include "annexb.h"
#include "parset.h"


/*!
 ************************************************************************
 *  \brief
 *     This function generates the slice (and partition) header(s)
 *
 *  \return number of bits used for the slice (and partition) header(s)
 *
 *  \par Side effects:
 *      Adds slice/partition header symbols to the symbol buffer
 *
 *  \par Remarks:
 *      Possible bugs:
 *      1. For some reason, the interlace folk has removed assert()s
 *         that check for byte alignment at the slice header position.
 *         Don't know why this is a problem, but it is surely against
 *         the spec to have slices starting at non byte-aligned positions
 ************************************************************************
*/

int start_slice()
{
  EncodingEnvironmentPtr eep;
  Slice *currSlice = img->currentSlice;
  Bitstream *currStream;
  int header_len = 0;
  int i;
  int NumberOfPartitions = (input->partition_mode == PAR_DP_1?1:3);

  assert (currSlice->max_part_nr == NumberOfPartitions);

  currSlice->num_mb = 0;          // no coded MBs so far
  RTPUpdateTimestamp (img->tr);   // this has no side effects, just leave it for all NALs

  for (i=0; i<NumberOfPartitions; i++)
  {
    currStream = (currSlice->partArr[i]).bitstream;
    assert (currStream != NULL);
    assert (currStream->bits_to_go == 8);
    assert (currStream->byte_buf == 0);
    assert (currStream->byte_pos == 0);
    memset(currStream->streamBuffer, 0, 12);    // fill first 12 bytes with zeros (debug only)
    currStream->last_startcode=currStream->byte_pos;
    currStream->write_flag = 0;     //! don't write this partition (flag will be set later when data is written into partition)
    if (i==0)     // First partition
      header_len += SliceHeader (0);
    else          // Second/Third partition
      header_len += Partition_BC_Header(i);
     
    //! Now the CABAC stuff...
    if (input->symbol_mode == CABAC)
    {
      eep = &((currSlice->partArr[i]).ee_cabac);
      if (currStream->bits_to_go != 8)
        header_len+=currStream->bits_to_go;
      ByteAlign (currStream);       //! Is this CABAC byte alignment still necessary???
      arienco_start_encoding(eep, currStream->streamBuffer, &(currStream->byte_pos),&(currStream->last_startcode),img->type);
      init_contexts_MotionInfo (currSlice->mot_ctx);
      init_contexts_TextureInfo(currSlice->tex_ctx);
    }
  }
  return header_len;
}


void writeUnit(Bitstream* currStream)
{
  NALU_t *nalu;
  assert (currStream->bits_to_go == 8);
  nalu = AllocNALU();
  nalu->startcodeprefix_len = 2+(img->current_mb_nr == 0?ZEROBYTES_SHORTSTARTCODE+1:ZEROBYTES_SHORTSTARTCODE);
  // printf ("nalu->startcodeprefix_len %d\n", nalu->startcodeprefix_len);
  nalu->len = currStream->byte_pos +1;            // add one for the first byte of the NALU
  printf ("nalu->len %d\n", nalu->len);
  memcpy (&nalu->buf[1], currStream->streamBuffer, nalu->len-1);
  if (img->idr_flag)
  {
    nalu->nal_unit_type = NALU_TYPE_IDR;
    nalu->nal_reference_idc = NALU_PRIORITY_HIGHEST;
  }
  else if (img->type == B_IMG)
  {
    nalu->nal_unit_type = NALU_TYPE_SLICE;
    nalu->nal_reference_idc = NALU_PRIORITY_DISPOSABLE;
    assert (img->disposable_flag == TRUE);
  }
  else   // non-disposable, non IDR slice
  {
    nalu->nal_unit_type = NALU_TYPE_SLICE;
    nalu->nal_reference_idc = NALU_PRIORITY_HIGH;
  }
  nalu->forbidden_bit = 0;
  stat->bit_ctr += WriteNALU (nalu) * 8;
  
  FreeNALU(nalu);

  currStream->stored_bits_to_go = 8; // store bits_to_go
  currStream->stored_byte_buf   = currStream->byte_buf;   // store current byte
  currStream->stored_byte_pos   = 0; // reset byte position

}

void writeVlcTrailingBits(Bitstream* currStream)
{
  if (currStream->bits_to_go < 8)
  { // trailing bits to process
    currStream->byte_buf <<= currStream->bits_to_go;
    stat->bit_use_stuffingBits[img->type]+=currStream->bits_to_go;
    currStream->streamBuffer[currStream->byte_pos++]=currStream->byte_buf;
    currStream->bits_to_go = 8;
  }
}


/*!
 ************************************************************************
 * \brief
 *    This function terminates a slice
 * \return
 *    0 if OK,                                                         \n
 *    1 in case of error
 *
 *
 * \par Remarks
 *    The RTP section of this function included the support for packet
 *    aggregation (for the SEI messages).  I tried to move all this to
 *    the RTP module.  Don't know whether I succeeded.  Nokia, please
 *    check.  Thanks.  StW, 12/08/02
 ************************************************************************
 */
int terminate_slice(int write_out)
{
  int bytes_written;
  Bitstream *currStream;
  Slice *currSlice = img->currentSlice;
  EncodingEnvironmentPtr eep;
  int i;
  int temp_byte_pos;
  NALU_t *nalu;

  if (input->symbol_mode == CABAC && !write_out)
    write_terminating_bit (1);      // only once, not for all partitions
  
  for (i=0; i<currSlice->max_part_nr; i++)
  {
    currStream = (currSlice->partArr[i]).bitstream;
    if (input->symbol_mode == UVLC)
    {
      if (img->fld_flag)
      {
        writeVlcTrailingBits((currSlice->partArr[i]).bitstream_topfld);
        writeVlcTrailingBits((currSlice->partArr[i]).bitstream_botfld);
      }
      else
      {
        writeVlcTrailingBits((currSlice->partArr[i]).bitstream_frm);
      }

      if(!write_out)
      {
        SODBtoRBSP(currStream);
        temp_byte_pos = currStream->byte_pos;
        currStream->byte_pos = RBSPtoEBSP(currStream->streamBuffer, currStream->last_startcode /*+startcodeprefix_len*/, currStream->byte_pos, 0);
        *(stat->em_prev_bits) += (currStream->byte_pos - temp_byte_pos) * 8;

        currStream->stored_bits_to_go = 8; // store bits_to_go
        currStream->stored_byte_buf   = currStream->byte_buf;   // store current byte
        currStream->stored_byte_pos   = currStream->byte_pos;   // reset byte position
      } 
      else
      {
        if (img->fld_flag)
        {
          writeUnit((currSlice->partArr[i]).bitstream_topfld);
          writeUnit((currSlice->partArr[i]).bitstream_botfld);
        }
        else
        {
          writeUnit((currSlice->partArr[i]).bitstream_frm);
        }
      }

    }
    else     // CABAC
    {
      if(!write_out)
      {
        write_terminating_bit (1);
        eep = &((currSlice->partArr[i]).ee_cabac);
        // terminate the arithmetic code
        arienco_done_encoding(eep);
        currStream->bits_to_go = eep->Ebits_to_go;
        currStream->byte_buf = 0;
        bytes_written = currStream->byte_pos-currStream->tmp_byte_pos;
         temp_byte_pos = currStream->byte_pos;
        currStream->byte_pos = RBSPtoEBSP(currStream->streamBuffer, currStream->last_startcode/*+startcodeprefix_len*/, currStream->byte_pos, eep->E);
        *(stat->em_prev_bits) += (currStream->byte_pos - temp_byte_pos) * 8;
        currStream->tmp_byte_pos = currStream->byte_pos;
      }
      else
      {
        nalu = AllocNALU();
        nalu->startcodeprefix_len = 2+(img->current_mb_nr == 0?ZEROBYTES_SHORTSTARTCODE+1:ZEROBYTES_SHORTSTARTCODE);
// printf ("nalu->startcodeprefix_len %d\n", nalu->startcodeprefix_len);
        nalu->len = currStream->byte_pos +1;            // add one for the first byte of the NALU
        memcpy (&nalu->buf[1], currStream->streamBuffer, nalu->len-1);
        if (img->idr_flag)
        {
           nalu->nal_unit_type = NALU_TYPE_IDR;
           nalu->nal_reference_idc = NALU_PRIORITY_HIGHEST;
        }
        else if (img->type == B_IMG)
        {
          nalu->nal_unit_type = NALU_TYPE_SLICE;
          nalu->nal_reference_idc = NALU_PRIORITY_DISPOSABLE;
          assert (img->disposable_flag == TRUE);
        }
        else   // non-disposable, non IDR slice
        {
          nalu->nal_unit_type = NALU_TYPE_SLICE;
          nalu->nal_reference_idc = NALU_PRIORITY_HIGH;
        }
         nalu->forbidden_bit = 0;
        stat->bit_ctr += WriteNALU (nalu) * 8;

        FreeNALU(nalu);
      }
    }           // CABAC
  }           // partition loop
/*
    case PAR_OF_RTP:
      printf ("RTP not implemented, commented code needs more work\n");
      if (input->symbol_mode==CABAC)
      {
        write_terminating_bit (1);
      }

      for (i=0; i<currSlice->max_part_nr; i++)
      {
        currStream = (currSlice->partArr[i]).bitstream;

        if (input->symbol_mode == CABAC)
        {
          eep = &((currSlice->partArr[i]).ee_cabac);
          arienco_done_encoding(eep);
                                        currStream->bits_to_go = eep->Ebits_to_go;
                                        currStream->byte_buf = 0;
                                }
        
        if (currStream->bits_to_go < 8) // trailing bits to process
        {
          currStream->byte_buf <<= currStream->bits_to_go;
          stat->bit_use_stuffingBits[img->type]+=currStream->bits_to_go;
          currStream->streamBuffer[currStream->byte_pos++]=currStream->byte_buf;
          currStream->bits_to_go = 8;
        }
        bytes_written = currStream->byte_pos; // number of written bytes
        
        if(img->type == B_IMG && i > 0)    // to be really sure
          currStream->write_flag = 0;

        if(currStream->write_flag == 1)                                                   
        {
          int Marker;
          int FirstBytePacketType, subFirstBytePacketType;
         
          // Tian Dong (Sept 2002):
          // The SEI message will be encapsulated together with the slice data into a compound
          // packet (the newly name is "aggregation packet", see draft-wenger-avt-rtp-jvt-01.txt)
          FirstBytePacketType = subFirstBytePacketType = 0;
          if (isAggregationPacket())  // has spare reference picture information to be packetized
          {
            FirstBytePacketType = AGGREGATION_PACKET_TYPE; //! compound packet (aggregation packet) See VCEG-N72
            if ( FirstFrameIn2ndIGOP == img->number )
              subFirstBytePacketType = (PAYLOAD_TYPE_IDERP << 4);
            if (input->partition_mode==PAR_DP_1 || img->type == B_IMG)
              subFirstBytePacketType |= 0;
            else
              subFirstBytePacketType |= i+1; //! See VCEG-N72
          }
          else 
          {
            if ( FirstFrameIn2ndIGOP == img->number )
              FirstBytePacketType = (PAYLOAD_TYPE_IDERP << 4);
            if ( input->partition_mode==PAR_DP_1 || img->type == B_IMG)
              FirstBytePacketType |= 0;
            else
              FirstBytePacketType |= i+1; //! See VCEG-N72
          }
          
          // Calculate the Marker bit
          LastPartition=0;
          if (input->partition_mode == PAR_DP_3 && img->type != B_IMG)
          {
            if (currSlice->partArr[1].bitstream->write_flag == 1)  // something in partition 1
              LastPartition=1;
            if (currSlice->partArr[2].bitstream->write_flag == 1)  // something in partition 2
              LastPartition=2;
          }
          Marker = 0;
          if (img->current_mb_nr == img->total_number_mb)   // last MB is in this slice
            if ((input->partition_mode==PAR_DP_1) ||          // Single SLice
                (input->partition_mode==PAR_DP_3 && i == LastPartition ))   // Last partition containing bits
              Marker = 1;

          // and write the RTP packet
          currStream->byte_pos = bytes_written; 
          currStream->bits_to_go = 8;
          currStream->byte_buf = 0;

                                        if (input->symbol_mode==UVLC)  //GB stopbit is inserted in arienco_done_encoding for CABAC
                                                SODBtoRBSP(currStream);

          bytes_written = currStream->byte_pos;
          temp_byte_pos = bytes_written;

          if (input->symbol_mode==CABAC) 
          {
                  eep = &((currSlice->partArr[i]).ee_cabac);
                  bytes_written = RBSPtoEBSP(currStream->streamBuffer, Bytes_After_Header, bytes_written,eep->E);
          }
                else
                  bytes_written = RBSPtoEBSP(currStream->streamBuffer, Bytes_After_Header, bytes_written,0);

          *(stat->em_prev_bits) += (bytes_written - temp_byte_pos) * 8;

          // Tian Dong (Sept 2002):
          if (isAggregationPacket())
            rtp_bytes_written = aggregationRTPWriteBits (Marker, FirstBytePacketType, subFirstBytePacketType, currStream->streamBuffer, bytes_written, out);
          else
            rtp_bytes_written = RTPWriteBits (Marker, FirstBytePacketType, currStream->streamBuffer, bytes_written, out);

          // JVT-D101, write the redundant slice
          if(input->redundant_slice_flag)
          {
            // seems that SODBtoRBSP() and RBSPtoEBSP() did not change bits of slice header, 
            // therefore a bit position in slice header is unchanged.
            img->redundant_pic_cnt = 1;
            modify_redundant_pic_cnt(currStream->streamBuffer);
            rtp_bytes_written = RTPWriteBits (Marker, FirstBytePacketType, currStream->streamBuffer, bytes_written, out);
          }
          // End JVT-D101
        }
        // stat->bit_ctr += 8*bytes_written;
        stat->bit_ctr += 8*bytes_written*(img->redundant_pic_cnt+1); // JVT-D101: modified from the above line

        // Provide the next partition with a 'fresh' buffer
        currStream->stored_bits_to_go = 8;
        currStream->stored_byte_buf   = 0;
        currStream->stored_byte_pos   = 0;
        currStream->bits_to_go = 8;
        currStream->byte_buf   = 0;
        currStream->byte_pos   = 0;
      }
    return 0;
*/
  return 0;   
}

// JVT-D101: redundant slices
/*!
 ************************************************************************
 * \brief
 *    This function set the value of a bit in a bitstream to 1
 ************************************************************************
 */
void modify_redundant_pic_cnt(unsigned char *buffer)
{
  unsigned char tmp = 1 << (rpc_bits_to_go-1);
  buffer[rpc_bytes_to_go] |= tmp;
}
// End JVT-D101



/*!
 *****************************************************************************
 *
 * \brief 
 *    int RTPSliceHeader () write the Slice header for the RTP NAL      
 *
 * \return
 *    Number of bits used by the slice header
 *
 * \para Parameters
 *    none
 *
 * \para Side effects
 *    Slice header as per VCEG-N72r2 is written into the appropriate partition
 *    bit buffer
 *
 * \para Limitations/Shortcomings/Tweaks
 *    The current code does not support the change of picture parameters within
 *    one coded sequence, hence there is only one parameter set necessary.  This
 *    is hard coded to zero.
 *    As per Email discussions on 10/24/01 we decided to include the SliceID
 *    into both the Full Slice pakcet and Partition A packet
 *   
 * \date
 *    October 24, 2001
 *
 * \author
 *    Stephan Wenger   stewe@cs.tu-berlin.de
 *****************************************************************************/

/*
int RTPSliceHeader()
{
  int dP_nr;
  Bitstream *currStream; 
  DataPartition *partition;
  SyntaxElement sym;

  int len = 0;
  int RTPSliceType;

  if(img->type==BS_IMG && img->type!=B_IMG)
    dP_nr= assignSE2partition[input->partition_mode][SE_BFRAME];
  else
    dP_nr= assignSE2partition[input->partition_mode][SE_HEADER];

  currStream = ((img->currentSlice)->partArr[dP_nr]).bitstream;
  partition = &((img->currentSlice)->partArr[dP_nr]);

  assert (input->of_mode == PAR_OF_RTP);
  sym.type = SE_HEADER;         // This will be true for all symbols generated here
  sym.mapping = n_linfo2;       // Mapping rule: Simple code number to len/info


  SYMTRACESTRING("RTP-SH: Parameter Set");
  sym.value1 = 0;
  len += writeSyntaxElement_UVLC (&sym, partition);
  
  // picture structure, (0: progressive, 1: top field, 2: bottom field, 
  // 3: top field first, 4: bottom field first)
  sym.value1 = img->pstruct;
  SYMTRACESTRING("Picture Stucture");
  len += writeSyntaxElement_UVLC (&sym, partition);

  SYMTRACESTRING("RTP-SH: Picture ID");
  sym.value1 = img->currentSlice->picture_id;

  len += writeSyntaxElement_UVLC (&sym, partition);

  switch (img->type)
  {
    case INTER_IMG:
      if (img->types==SP_IMG)
        RTPSliceType = 2;
      else
        RTPSliceType = 0;
      break;
    case INTRA_IMG:
      RTPSliceType = 3;
      break;
    case B_IMG:
    case BS_IMG:
      RTPSliceType = 1;
      break;
    case SP_IMG:
      // That's what I would expect to be the case. But img->type contains INTER_IMG 
      // here (see handling above)
      //! \todo make one variable from img->type and img->types
      RTPSliceType = 2;
      break;
    default:
      printf ("Panic: unknown picture type %d, exit\n", img->type);
      exit (-1);
  }


  SYMTRACESTRING("RTP-SH: Slice Type");
  sym.value1 = RTPSliceType;
  len += writeSyntaxElement_UVLC (&sym, partition);

  // NOTE: following syntax elements have to be moved into nal_unit() and changed from e(v) to u(1)
  if(1)
  {
    sym.value1 = img->disposable_flag;
    SYMTRACESTRING("RTP-SH: NAL disposable_flag");
    len += writeSyntaxElement_UVLC (&sym, partition);
  }
  
  // NOTE: following syntax elements have to be moved into picture_layer_rbsp()
  // weighted_bipred_implicit_flag
  if(img->type==INTER_IMG)
  {
    sym.value1 = input->WeightedPrediction;
    SYMTRACESTRING("PH weighted_pred_flag");
    len += writeSyntaxElement_UVLC (&sym, partition);
  }

  if(img->type==B_IMG || img->type==BS_IMG)
  {
    sym.value1 = (input->WeightedBiprediction == 1)? 1 : 0;
    SYMTRACESTRING("PH weighted_bipred_explicit_flag");
    len += writeSyntaxElement_UVLC (&sym, partition);
  }
  if(img->type==B_IMG || img->type==BS_IMG)
  {
    sym.value1 = (input->WeightedBiprediction == 2)? 1 : 0;
    SYMTRACESTRING("PH weighted_bipred_implicit_flag");
    len += writeSyntaxElement_UVLC (&sym, partition);
  }

  if (img->type==INTER_IMG || img->type==B_IMG || img->type==BS_IMG)
  {
    sym.value1 = img->num_ref_pic_active_fwd_minus1;
    SYMTRACESTRING("RTP-SH: num_ref_pic_active_fwd_minus1");
    len += writeSyntaxElement_UVLC (&sym, partition);
    if (img->type==B_IMG || img->type==BS_IMG)
    {
      sym.value1 = img->num_ref_pic_active_bwd_minus1;
      SYMTRACESTRING("RTP-SH: num_ref_pic_active_bwd_minus1");
      len += writeSyntaxElement_UVLC (&sym, partition);
    }
  }

  SYMTRACESTRING("RTP-SH: FirstMBInSliceX");
  sym.value1 = img->mb_x;
  len += writeSyntaxElement_UVLC (&sym, partition);

  SYMTRACESTRING("RTP-SH: FirstMBinSlinceY");
  sym.value1 = img->mb_y;
  len += writeSyntaxElement_UVLC (&sym, partition);

  // JVT-D101: redundant_pic_cnt should be a ue(v) instead of u(1). However since currently we allow 
  // only 1 redundant slice for each slice, therefore herein encoding it as u(1) is OK.
  if(input->redundant_slice_flag) 
  {
    SYMTRACESTRING("RTP-SH: redundant_pic_cnt");
    rpc_bytes_to_go = currStream->byte_pos;
    rpc_bits_to_go = currStream->bits_to_go;
    sym.bitpattern = img->redundant_pic_cnt = 0; //may be modifed before being written into bitstream file
    sym.len = 1;
    len += writeSyntaxElement_fixed(&sym, partition);
  }
  // End JVT-D101

  if (img->type==B_IMG || img->type==BS_IMG)
  {
    SYMTRACESTRING("RTP-SH DirectSpatialFlag");
    sym.bitpattern = input->direct_type;  // 1 for Spatial Direct, 0 for temporal Direct
    sym.len = 1;
    len += writeSyntaxElement_fixed(&sym, partition);
  }

#ifdef _ABT_FLAG_IN_SLICE_HEADER_
  SYMTRACESTRING("RTP-SH ABTMode");
  sym.value1 = input->abt;
  len += writeSyntaxElement_UVLC (&sym, partition);
#endif

  SYMTRACESTRING("RTP-SH: InitialQP");
  sym.mapping = dquant_linfo;
  sym.value1 = img->qp - (MAX_QP - MIN_QP +1)/2;
  len += writeSyntaxElement_UVLC (&sym, partition);

  if (img->types==SP_IMG)
  {

    if (img->type!=INTRA_IMG) // Switch Flag only for SP pictures
    {
      SYMTRACESTRING("RTP-SH SWITCH FLAG");
      sym.bitpattern = 0;  // 1 for switching SP, 0 for normal SP
      sym.len = 1;
      len += writeSyntaxElement_fixed(&sym, partition);
    }

    SYMTRACESTRING("RTP-SH: SP InitialQP");
    sym.value1 = img->qpsp - (MAX_QP - MIN_QP +1)/2;
    len += writeSyntaxElement_UVLC (&sym, partition);
  }

  if (input->LFSendParameters)
  {
    SYMTRACESTRING("RTP-SH LF_DISABLE FLAG");
    sym.bitpattern = input->LFDisable;  
    sym.len = 1;
    len += writeSyntaxElement_fixed(&sym, partition);

    if (!input->LFDisable)
    {
      sym.mapping = dquant_linfo;           // Mapping rule: Signed integer
      SYMTRACESTRING("RTP-SH LFAlphaC0OffsetDiv2");
      sym.value1 = input->LFAlphaC0Offset>>1; 
      len += writeSyntaxElement_UVLC (&sym, partition);

      SYMTRACESTRING("RTP-SH LFBetaOffsetDiv2");
      sym.value1 = input->LFBetaOffset>>1; 
      len += writeSyntaxElement_UVLC (&sym, partition);
    }
  }

  sym.mapping = n_linfo2;


  if (input->partition_mode == PAR_DP_3 && img->type != B_IMG && img->type != BS_IMG)
  {
    SYMTRACESTRING("RTP-SH: Slice ID");
    sym.value1 = img->current_slice_nr;
    len += writeSyntaxElement_UVLC (&sym, partition);
  }

  len += writeERPS(&sym, partition);

  // JVT-D097
  if( input->num_slice_groups_minus1 > 0  &&  
      input->mb_allocation_map_type >= 4  &&  
                  input->mb_allocation_map_type <= 6)
  {
    SYMTRACESTRING("RTP-SH: slice_group_change_cycle");
    sym.value1 = slice_group_change_cycle;
    len += writeSyntaxElement_UVLC (&sym, partition);
  }
  // End JVT-D097

  // After this, and when in CABAC mode, terminateSlice() may insert one more
  // UVLC codeword for the number of coded MBs

  return len;
}
*/